/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: LibvirtDomainConfig.java 103 2011-07-04 15:53:39Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.libvirt;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.main.AgentControl;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

public class LibvirtDomainConfig {
    static Logger logger = Logger.getLogger(LibvirtDomainConfig.class);

    // static {
    // try {
    // LibvirtDomainConfig.newDomainConfig("kvm.tmpl");
    // } catch (Exception ex) {
    // LibvirtDomainConfig.logger.error("Failed to load KVM libvirt XML template file",
    // ex);
    // }
    // }

    private Document xmlDocument;

    private Node devicesNode;

    private NodeList diskNodeTemplates;

    private NodeList vnicNodeTemplates;

    private String diskDevicePrefix, cdromDevicePrefix;

    private Node diskNodeTemplate;

    private Node cdromNodeTemplate;

    private Node vnicNodeTemplate;

    private LibvirtDomainConfig(final Document xmlDocument, final boolean isTemplate) {
        this.xmlDocument = xmlDocument;
        this.devicesNode = XMLUtils.getNodeByPath(xmlDocument, "/domain/devices").item(0);
        this.diskNodeTemplates = XMLUtils.getNodeByPath(xmlDocument, "/domain/devices/disk");
        if (isTemplate) {
            for (int i = 0; i < this.diskNodeTemplates.getLength(); i++) {
                Node parent = this.diskNodeTemplates.item(i).getParentNode();
                parent.removeChild(this.diskNodeTemplates.item(i));
                Node n = this.diskNodeTemplates.item(i).getAttributes().getNamedItem("device");
                if (n != null) {
                    String deviceType = ((Attr) n).getValue();
                    if (deviceType.equals("disk")) {
                        this.diskNodeTemplate = this.diskNodeTemplates.item(i);
                        this.diskDevicePrefix = XMLUtils.getAttributeValue(this.diskNodeTemplate, "target", "dev").substring(0,
                            2);
                    } else if (deviceType.equals("cdrom")) {
                        this.cdromNodeTemplate = this.diskNodeTemplates.item(i);
                        this.cdromDevicePrefix = XMLUtils.getAttributeValue(this.cdromNodeTemplate, "target", "dev").substring(
                            0, 2);
                    }
                }
            }
        }
        this.vnicNodeTemplates = XMLUtils.getNodeByPath(xmlDocument, "/domain/devices/interface");
        if (isTemplate) {
            for (int i = 0; i < this.vnicNodeTemplates.getLength(); i++) {
                Node parent = this.vnicNodeTemplates.item(i).getParentNode();
                parent.removeChild(this.vnicNodeTemplates.item(i));
                this.vnicNodeTemplate = this.vnicNodeTemplates.item(i);
            }
        }
    }

    public static LibvirtDomainConfig newDomainConfig(final String templateFileName) throws Exception {
        String templatePath = AgentControl.vmmHome + File.separator + "etc" + File.separator + "libvirt" + File.separator
            + templateFileName;
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        Document xmlDocument;

        DocumentBuilder db = dbf.newDocumentBuilder();
        xmlDocument = db.parse(new File(templatePath));
        return new LibvirtDomainConfig(xmlDocument, true);
    }

    public static LibvirtDomainConfig loadDomainConfig(final String domainXMLConfig) throws Exception {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        Document xmlDocument;

        DocumentBuilder db = dbf.newDocumentBuilder();
        xmlDocument = db.parse(new ByteArrayInputStream(domainXMLConfig.getBytes("UTF-8")));
        return new LibvirtDomainConfig(xmlDocument, false);
    }

    public List<Node> getDiskNodes(final String device, final String type) {
        List<Node> result = new ArrayList<Node>();
        for (int i = 0; i < this.diskNodeTemplates.getLength(); i++) {
            Node deviceAttr = this.diskNodeTemplates.item(i).getAttributes().getNamedItem("device");
            Node typeAttr = this.diskNodeTemplates.item(i).getAttributes().getNamedItem("type");
            if (deviceAttr != null && ((Attr) deviceAttr).getValue().equals(device) && typeAttr != null
                && ((Attr) typeAttr).getValue().equals(type)) {
                result.add(this.diskNodeTemplates.item(i));
            }
        }
        return result;
    }

    public String getDiskDevicePrefix() {
        return this.diskDevicePrefix;
    }

    public String getCdromDevicePrefix() {
        return this.cdromDevicePrefix;
    }

    public boolean updateCdromXmlDefinition(final String deviceName, final String isoFilePath) {
        for (int i = 0; i < this.diskNodeTemplates.getLength(); i++) {
            Node parent = this.diskNodeTemplates.item(i).getParentNode();
            Node n = this.diskNodeTemplates.item(i).getAttributes().getNamedItem("device");
            if (n != null) {
                String deviceType = ((Attr) n).getValue();
                if (deviceType.equals("cdrom")) {
                    Node cdromNode = this.diskNodeTemplates.item(i);
                    if (XMLUtils.getAttributeValue(cdromNode, "target", "dev").equals(deviceName)) {
                        this.updateAttributeValue(cdromNode, "source", "file", isoFilePath);
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public void setName(final String name) {
        this.updateNodeText("name", name);
    }

    public void updateNodeText(final String nodeName, final String nodeText) {
        Node node = this.xmlDocument.getDocumentElement().getFirstChild();
        while (node != null) {
            if (node.getNodeName().equals(nodeName)) {
                break;
            }
            node = node.getNextSibling();
        }
        if (node != null) {
            Text txt = (Text) node.getFirstChild();
            txt.replaceWholeText(nodeText);
        }
    }

    public void setUUID(final String uuid) {
        this.updateNodeText("uuid", uuid);
    }

    public void setNumVCPUs(final int numVCPUs) {
        this.updateNodeText("vcpu", Integer.toString(numVCPUs));
    }

    public void setMemoryMB(final int memoryMB) {
        this.updateNodeText("memory", Integer.toString(memoryMB * 1024));
    }

    public void setDescription(final String description) {
        this.updateNodeText("description", description);
    }

    public void setBootDevice(final String bootDevice) {
        NodeList nl = XMLUtils.getNodeByPath(this.xmlDocument, "/domain/os");
        if (nl.getLength() == 1) {
            Node bootNode = nl.item(0);
            this.updateAttributeValue(bootNode, "boot", "dev", bootDevice);
        }
    }

    public void addDisk(final String imageFilePath, final String targetDev) {
        Node newDisk = this.diskNodeTemplate.cloneNode(true);
        this.updateAttributeValue(newDisk, "source", "file", imageFilePath);
        this.updateAttributeValue(newDisk, "target", "dev", targetDev);
        this.devicesNode.appendChild(newDisk);
    }

    public String getDiskDesc(final String imageFilePath, final String targetDev, final String format) {
        Node newDisk = this.diskNodeTemplate.cloneNode(true);
        this.updateAttributeValue(newDisk, "source", "file", imageFilePath);
        this.updateAttributeValue(newDisk, "target", "dev", targetDev);
        this.updateAttributeValue(newDisk, "driver", "type", format);
        return XMLUtils.xmlToString(newDisk);
    }

    public void addCdrom(final String imageFilePath, final String targetDev) {
        // Node n = LibvirtDomainConfig.cdromNodeTemplate.cloneNode(true);
        Node newCdrom = this.xmlDocument.importNode(this.cdromNodeTemplate, true);
        this.updateAttributeValue(newCdrom, "source", "file", imageFilePath);
        this.updateAttributeValue(newCdrom, "target", "dev", targetDev);
        this.devicesNode.appendChild(newCdrom);
    }

    public void addVNIC(final String macAddress) {
        Node newVNIC = this.vnicNodeTemplate.cloneNode(true);
        if (macAddress != null) {
            Node macNode = this.xmlDocument.createElement("mac");
            Attr macAddressAttr = this.xmlDocument.createAttribute("address");
            macAddressAttr.setValue(macAddress);
            macNode.getAttributes().setNamedItem(macAddressAttr);
            newVNIC.appendChild(macNode);
        }
        this.devicesNode.appendChild(newVNIC);
    }

    public void updateAttributeValue(final Node parent, final String nodeName, final String attributeName,
        final String attributeValue) {
        Node child = parent.getFirstChild();
        while (child != null) {
            if (child.getNodeName().equals(nodeName)) {
                break;
            }
            child = child.getNextSibling();
        }
        if (child != null) {
            Attr sourceAttr = (Attr) child.getAttributes().getNamedItem(attributeName);
            if (sourceAttr != null) {
                sourceAttr.setValue(attributeValue);
            } else {
                sourceAttr = this.xmlDocument.createAttribute(attributeName);
                sourceAttr.setValue(attributeValue);
                child.getAttributes().setNamedItem(sourceAttr);
            }
        } else {
            Node node = this.xmlDocument.createElement(nodeName);
            Attr attr = this.xmlDocument.createAttribute(attributeName);
            attr.setValue(attributeValue);
            node.getAttributes().setNamedItem(attr);
            parent.appendChild(node);
        }
    }

    @Override
    public String toString() {
        return XMLUtils.xmlToString(this.xmlDocument);
    }

}
